# Copyright 1999-2017. Parallels IP Holdings GmbH. All Rights Reserved.

import sys
import os
import json

from optparse import OptionParser
import plesk_subprocess as subprocess
import plesk_log

USAGE = '''
  %prog [options] <command> [command-arguments]

Commands:
  versions              display the versions of NodeJS instances known to
                        Plesk
  register <path-to-node>
                        make the specified version of NodeJS instance
                        known to Plesk
  unregister <path-to-node>
                        make the specified version of NodeJS instance
                        unknown to Plesk
'''

NODE_VERSIONS_DIR = '/etc/psa/node_versions.d/'
NODE_VERSIONS_LIST = '/var/lib/plesk/node_versions.json'

log = plesk_log.getLogger(__name__)


class Node(object):

    def __init__(self, **kwargs):
        for keyword in ["path"]:
            setattr(self, keyword, kwargs[keyword])

        try:
            self.version = subprocess.check_output([self.path, '--version']).strip()
        except Exception as e:
            raise type(e)('cannot get version for %s: %s' %
                          (self.path, str(e)))

    def _to_string(self):
        return "%s %s" % (self.version, self.path)

    def _to_dict(self):
        output = {}
        for k in ['version', 'path']:
            output[k] = getattr(self, k)
        return output

    def __eq__(self, othr):
        return (isinstance(othr, type(self)) and (self.path == othr.path))

    def __hash__(self):
        return hash(self.path)

class NodeJsonEncoder(json.JSONEncoder):

    def default(self, obj):
        if isinstance(obj, Node):
            obj = obj._to_dict()
        elif isinstance(obj, set):
            obj = list(obj)
        else:
            obj = super(NodeJsonEncoder, self).default(obj)
        return obj


def _get_permanent_versions():
    versions = set()

    if not os.path.isdir(NODE_VERSIONS_DIR):
        return versions

    for entry in os.listdir(NODE_VERSIONS_DIR):
        filename = os.path.join(NODE_VERSIONS_DIR, entry)
        if not os.path.isfile(filename):
            continue

        try:
            with open(filename, 'r') as descriptor:
                version_info = json.load(descriptor)
                r = Node(**version_info)

                if r in versions:
                    log.warn('duplicate version %s', r.version)

                versions.add(r)

        except Exception as e:
            log.warn('could not process file %s: %s', filename, e)

    return versions


def _get_custom_versions():
    versions = set()

    if os.path.isfile(NODE_VERSIONS_LIST):
        try:
            with open(NODE_VERSIONS_LIST, 'r') as input_file:
                for version_info in json.load(input_file):
                    try:
                        r = Node(**version_info)
                    except Exception as e:
                        log.warn('Bad NodeJS inpterpreter %s', e)
                        continue
                        
                    versions.add(r)
        except Exception as e:
            log.warn('could not load custom versions file ' +
                     NODE_VERSIONS_LIST + ': ' + str(e))
    return versions


def _set_custom_versions(versions):
    with open(NODE_VERSIONS_LIST, 'w') as output_file:
        json.dump(versions, output_file, cls=NodeJsonEncoder)


def _get_known_versions():
    versions = _get_custom_versions()
    try:
        versions.update(_get_permanent_versions())
    except OSError:
        pass
    return versions


def _show_versions(output_format):
    versions = _get_known_versions()

    if output_format == 'json':
        json.dump(versions, sys.stdout, cls=NodeJsonEncoder)
        return

    if output_format == 'plain':
        for v in versions:
            print v._to_string()
        return


def _set_version(path):
    n = Node(path=path) 
    if n in _get_permanent_versions():
        raise Exception(
            'the specified version has been registered by a package; it can be updated with editing the corresponding file in ' + NODE_VERSIONS_DIR)

    versions = _get_custom_versions()
    versions.add(n)
    _set_custom_versions(versions)


def _discard_version(path):
    n = Node(path=path)
    if n in _get_permanent_versions():
        raise Exception('the specified version has been registered by a package; it can be unregistered with either uninstalling the package or removing the corresponding file from ' + NODE_VERSIONS_DIR)

    versions = _get_custom_versions()
    if n not in versions:
        return

    versions.discard(n)
    _set_custom_versions(versions)


def doMain():
    plesk_log.setDefaultLevel('warning')

    option_parser = OptionParser(usage=USAGE)
    option_parser.description = 'A utility for managing versions of NodeJS in Plesk'
    option_parser.add_option('-f', '--output-format', metavar='format',
                             type='choice', choices=['plain', 'json'], default='plain',
                             help='output info in the specified format (plain/json)')

    (options, arguments) = option_parser.parse_args()
    if not arguments:
        option_parser.error('missing command')

    command = arguments[0]
    arguments = arguments[1:]

    if command == 'versions':
        _show_versions(options.output_format)
    elif command == 'register':
        if len(arguments) != 1:
            option_parser.error('inappropriate number of command arguments')
        _set_version(arguments[0])
    elif command == 'unregister':
        if len(arguments) != 1:
            option_parser.error('inappropriate number of command arguments')
        _discard_version(arguments[0])
    else:
        option_parser.error('command `' + command + '` not recognized')


def main():
    try:
        doMain()
    except SystemExit:
        raise
    except Exception as e:
        log.error(e)
        log.debug(e, exc_info=sys.exc_info())
        sys.exit(1)

if __name__ == '__main__':
    main()

# vim: ts=4 sts=4 sw=4 et :
