# Copyright 1999-2017. Parallels IP Holdings GmbH. All Rights Reserved.
from optparse import OptionParser
import sys
import logging
import os.path
import plesk_subprocess as subprocess

from plesk_constants import PRODUCT_ROOT_D


class AptHelper:
    APTGET_ENV = {
        "DEBIAN_FRONTEND": "noninteractive",
        "LANG": "C",
        "PATH": "/usr/sbin:/usr/bin:/sbin:/bin",
    }
    APTGET_CMD = ('apt-get -qq --assume-yes '
                  '-o Dpkg::Options::=--force-confdef -o Dpkg::Options::=--force-confold '
                  '-o APT::Install-Recommends=no --purge').split()

    @classmethod
    def install(cls, packages):
        subprocess.check_call(["apt-get", "-qq", "update"], env=cls.APTGET_ENV)
        install_cmd = cls.APTGET_CMD + ['install'] + packages
        subprocess.check_call(install_cmd, env=cls.APTGET_ENV)

    @classmethod
    def remove(cls, packages):
        installed_packages = []
        with open(os.devnull, "wb") as devnull:
            for p in packages:
                try:
                    subprocess.check_call(["dpkg", "--status", p], stdout=devnull, stderr=devnull)
                    installed_packages.append(p)
                except subprocess.CalledProcessError as ex:
                    logging.debug(ex, exc_info=True)
                    logging.warning('Package "%s" is not installed', p)
        if installed_packages:
            remove_cmd = cls.APTGET_CMD + ['remove'] + installed_packages
            subprocess.check_call(remove_cmd, env=cls.APTGET_ENV)

    @classmethod
    def version(cls, packages):
        if len(packages) != 1:
            raise Exception("Exactly one package name should be specified")
        package = packages[0]
        cmd_status = ['dpkg-query', '--show', '--showformat', '${Status}', package]
        status = subprocess.check_output(cmd_status)
        if status != 'install ok installed':
            raise Exception("Package '%s' is not properly installed (%s)" % (package, status))
        cmd = ['dpkg-query', '--show', '--showformat', '${Version}\\n', package]
        subprocess.check_call(cmd)


class YumHelper:
    @classmethod
    def install(cls, packages):
        # accept arguments in format name=version:
        packages = [x.replace('=', '-', 1) for x in packages]
        cls._check_packages_available(packages)
        install_cmd = 'yum -y -q install'.split() + packages
        subprocess.check_call(install_cmd)

    @classmethod
    def _check_packages_available(cls, packages):
        with open(os.devnull, "wb") as devnull:
            for p in packages:
                try:
                    subprocess.check_call(['yum', '-q', 'info', p], stdout=devnull, stderr=devnull)
                except subprocess.CalledProcessError as ex:
                    logging.debug(ex, exc_info=True)
                    raise Exception('Package "%s" is not available' % p)

    @classmethod
    def remove(cls, packages):
        remove_cmd = 'yum -y -q remove'.split() + packages
        subprocess.check_call(remove_cmd)

    @classmethod
    def version(cls, packages):
        if len(packages) != 1:
            raise Exception("Exactly one package name should be specified")
        package = packages[0]
        cmd_version_release = ['rpm', '-q', '--queryformat', '%{Version}-%{Release}', package]
        version_release = subprocess.check_output(cmd_version_release)
        cmd_epoch = ['rpm', '-q', '--queryformat', '%{Epoch}', package]
        epoch = subprocess.check_output(cmd_epoch)
        full_version = (epoch + ":" + version_release) if epoch != '(none)' else version_release
        print(full_version)


def get_pkg_type():
    cmd = [os.path.join(PRODUCT_ROOT_D, "admin/sbin/osdetect"), "--pkg"]
    return subprocess.check_output(cmd).strip()


def main():
    try:
        parser = OptionParser()
        parser.add_option("--install", action="store_true", default=False, help="install specified packages")
        parser.add_option("--remove", action="store_true", default=False, help="remove specified packages")
        parser.add_option("--version", action="store_true", default=False, help="show version of specified package")
        (opts, args) = parser.parse_args()

        pkg = get_pkg_type()
        if pkg == "deb":
            helper = AptHelper()
        elif pkg == "rpm":
            helper = YumHelper()
        else:
            raise Exception("Unknown package type: '%s'" % pkg)

        if len(filter(None, [opts.install, opts.remove, opts.version])) != 1:
            raise Exception("One of --install, --remove, --version should be specified")

        if opts.install:
            helper.install(args)
        elif opts.remove:
            helper.remove(args)
        elif opts.version:
            helper.version(args)
        else:
            raise Exception("Unknown operation")

    except Exception as e:
        logging.error(e)
        logging.debug(e, exc_info=True)
        sys.exit(1)


if __name__ == "__main__":
    main()
